--
-- GPS with steering assistance
-- 
-- V2.0 MP ready and new features
-- release date 02.06.2013

-- upsideDown 
-- V1.0: 04/2013


GPS = {};
local GPS_directory = g_currentModDirectory;

source(Utils.getFilename("GPS_Event.lua", GPS_directory));
source(Utils.getFilename("GPS_SaveEvent.lua", GPS_directory));

function GPS:prerequisitesPresent(specializations)
    --return true;
	return SpecializationUtil.hasSpecialization(steerable, specializations);
end;

function GPS:load(xmlFile)
	
	if Steerable.GPSinserted == nil then
		Steerable.updateVehiclePhysics = Utils.overwrittenFunction(
			Steerable.updateVehiclePhysics,
			GPS.updateVehiclePhysics);
			
		Steerable.GPSinserted = true;
		print("GPS mod inserted into Steerable Specialization (should happen only once!)")
	end

	
	self.GPSWidth = 6;
	
	self.GPSfirstrun = true;
	self.lhBeta = 0;
	self.lhAngle = 0;
	self.Active = false;
	self.GPSshowTime = 0;
	
	self.GPSisActiveSteering = false;
	self.GPSangleModeLS13 = true;
	self.GPSclickTimer = 0;
	
	self.GPSshowMode = 0;
	
	self.GPSoverlay = Overlay:new("GPSoverlay", Utils.getFilename("GPShud.dds", GPS_directory),  0.3, 0.01238, 0.10, 0.04);
	
	self.GPS_LRoffset = 0;
	self.GPS_LRoffset_resetTimer = 0;
	
	self.GPStxt = {};
	self.GPStxt.txt = "";
	self.GPStxt.r = .0;
	self.GPStxt.g = 0.;
	self.GPStxt.b = 0.;
	self.GPStxt.bold = false;
	
	self.GPStxt.printDistance = 0;
	
	self.GPS_blinkTime  = 0;
	self.GPS_lastActionText = "";
end;



function GPS:updateVehiclePhysics(oldFunc, axisForward, axisForwardIsAnalog, axisSide, axisSideIsAnalog, dt)	
	local offset = 0;
	if self.GPSsteeringOffset ~= nil then
		offset = self.GPSsteeringOffset
	end

	oldFunc(self, axisForward, axisForwardIsAnalog, (axisSide+Utils.getNoNil(offset,0)), axisSideIsAnalog, dt);

end;


function GPS:mouseEvent(posX, posY, isDown, isUp, button)
end;

function GPS:keyEvent(unicode, sym, modifier, isDown)
end;

function GPS:updateTick(dt)
	self.GPSclickTimer = self.GPSclickTimer - dt;
	if self.GPSclickTimer < 0 then
		self.GPSclickTimer = 0;
	end
	self.GPS_LRoffset_resetTimer = self.GPS_LRoffset_resetTimer - dt
	if self.GPS_LRoffset_resetTimer < 0 then
		self.GPS_LRoffset_resetTimer = 0;
	end
	
	if self.GPSActive then
		if self.GPSisActiveSteering then
			self.GPStxt.txt = Steerable.GPS_TXT_ACT;
			self.GPStxt.r = .0;
			self.GPStxt.g = 0.6;
			self.GPStxt.b = 0;
			self.GPStxt.bold = true;
		else
			self.GPStxt.txt = Steerable.GPS_TXT_PAS;
			self.GPStxt.r = .0;
			self.GPStxt.g = 0.;
			self.GPStxt.b = 0.8;
			self.GPStxt.bold = false;
		end
		if self.GPSshowMode > 0 then
			self.GPStxt.txt = string.format("%2.2f m", self.GPStxt.printDistance)
		end
	else
		self.GPStxt.txt = Steerable.GPS_TXT_OFF;
		self.GPStxt.r = .7;
		self.GPStxt.g = 0;
		self.GPStxt.b = 0;
		self.GPStxt.bold = false;
	end

	if self.GPS_blinkTime > 0 then
		self.GPS_blinkTime = self.GPS_blinkTime - dt;
	else
		self.GPS_blinkTime = 0;
	end
	
	if self.GPSshowTime > 0 then
		self.GPSshowTime = self.GPSshowTime - dt;
	else
		self.GPSshowTime = 0;
	end
	
	if math.abs(self.axisSide) > 0.5 then --reset on manual steering
		self.GPSisActiveSteering = false;	
	end
	
	
	self.axisSide = self.axisSide + Utils.getNoNil(self.GPSsteeringOffset,0); --MP fix. Thats all folks ;-)
	if self.GPSActive and self.GPSisActiveSteering then
		self.axisSideIsAnalog = true;
	end -- end MP fix. It's all piggybacked on Steerable, no extra network traffic :-)
	
	
end;

function GPS:update(dt)

	local needPushEvent = false;
	
	if self:getIsActiveForInput(false) then
		
		if InputBinding.hasEvent(InputBinding.GPS_OnOff) then
			self.GPSActive = not self.GPSActive;
		end;
		
		
		if InputBinding.isPressed(InputBinding.GPS_OffsetLeft) and InputBinding.isPressed(InputBinding.GPS_OffsetRight) then
			self.GPS_LRoffset = 0;
			self.GPS_LRoffset_resetTimer = 800;
			needPushEvent = true;
		elseif self.GPS_LRoffset_resetTimer == 0 then
			if InputBinding.isPressed(InputBinding.GPS_OffsetRight) then
				self.GPS_LRoffset = self.GPS_LRoffset + .0008*dt;				
				self.GPSshowTime = 800;
				needPushEvent = true;
			end;
			
			if InputBinding.isPressed(InputBinding.GPS_OffsetLeft) then
				self.GPS_LRoffset = self.GPS_LRoffset - .0008*dt;				
				self.GPSshowTime = 800;
				needPushEvent = true;
			end;		
		end
		if InputBinding.isPressed(InputBinding.GPS_WidthPlus) then
					self.GPSWidth = self.GPSWidth + .0004*dt*(self.GPSWidth+1);
			if self.GPSWidth > 50 then
				self.GPSWidth = 50;
			end;
			self.GPSshowTime = 800;
			needPushEvent = true;
		end;
		
		if InputBinding.isPressed(InputBinding.GPS_WidthMinus) then
		
			self.GPSWidth = self.GPSWidth - .0004*dt*(self.GPSWidth+1);
			if self.GPSWidth < 0 then
				self.GPSWidth = 0;
			end;
			self.GPSshowTime = 800;
			needPushEvent = true;
		end;
		
		if self.GPS_LRoffset > self.GPSWidth/2 then
			self.GPS_LRoffset = self.GPSWidth/2;
		end;
		if self.GPS_LRoffset < -self.GPSWidth/2 then
			self.GPS_LRoffset = -self.GPSWidth/2;
		end;
		
		if InputBinding.hasEvent(InputBinding.GPS_SteeringOnOff) then
			self.GPSisActiveSteering = not self.GPSisActiveSteering;
		end
		
		if InputBinding.hasEvent(InputBinding.GPS_LS13LS11) then
			self.GPSangleModeLS13 = not self.GPSangleModeLS13;
		end
		
		if InputBinding.hasEvent(InputBinding.GPS_InvertOffset) then
			self.GPS_LRoffset = - self.GPS_LRoffset;
			self.GPSshowTime = 800;
			needPushEvent = true;
		end

		if InputBinding.hasEvent(InputBinding.GPS_InfoMode) then
			self.GPSshowMode = self.GPSshowMode + 1;
			if self.GPSshowMode > 3 then
				self.GPSshowMode = 0;
			end;
		end
		
		if InputBinding.hasEvent(InputBinding.GPS_Save) then
			if not self.GPSfirstrun then
				GPS.Store = {};
				GPS.Store.lhdX0 = self.lhdX0;
				GPS.Store.lhdZ0 = self.lhdZ0;
				GPS.Store.lhX0 = self.lhX0;
				GPS.Store.lhZ0 = self.lhZ0;
				GPS.Store.GPSWidth = self.GPSWidth;
				GPS.Store.GPS_LRoffset = self.GPS_LRoffset
				self.GPS_blinkTime = 4000;
				self.GPS_lastActionText = Steerable.GPS_TXT_SAVE
		
				if g_server ~= nil then
					g_server:broadcastEvent(GPS_SaveEvent:new(self, self.lhX0,self.lhZ0,self.lhdX0,self.lhdZ0,self.GPSWidth,self.GPS_LRoffset), nil, nil, self);
				else
					g_client:getServerConnection():sendEvent(GPS_SaveEvent:new(self, self.lhX0,self.lhZ0,self.lhdX0,self.lhdZ0,self.GPSWidth,self.GPS_LRoffset));
				end;	

				
			end			
		end

		if InputBinding.hasEvent(InputBinding.GPS_Load) then
			if GPS.Store ~= nil then
				self.GPSfirstrun = false
				self.lhdX0 = GPS.Store.lhdX0;
				self.lhdZ0 = GPS.Store.lhdZ0;
				self.lhX0 = GPS.Store.lhX0;
				self.lhZ0 = GPS.Store.lhZ0;
				self.GPSWidth = GPS.Store.GPSWidth;
				self.GPS_LRoffset = GPS.Store.GPS_LRoffset
				self.GPS_blinkTime = 2500;
				self.GPS_lastActionText = Steerable.GPS_TXT_LOAD
			end			
		end
		
		
		if InputBinding.hasEvent(InputBinding.GPS_NearestSteerable) then 
			local nVehicles = table.getn(g_currentMission.steerables);
			local nearestFoundIndex = 0
			local nearestFoundDistance = 100000;
			if nVehicles > 0 then
				local self_x,_,self_z = getWorldTranslation(self.rootNode)
				for i=1,nVehicles do
					local oVehicle = g_currentMission.steerables[i];
					local near_x,_,near_z = getWorldTranslation(oVehicle.rootNode)
					local nearestDistance = Utils.vector2Length(near_x-self_x,near_z-self_z);
					if oVehicle ~= self then
						if nearestDistance < nearestFoundDistance then
							nearestFoundDistance = nearestDistance;
							nearestFoundIndex = i;
						end
					end
				end				
				--here we have nearestFoundIndex, nearestFoundDistance
				if nearestFoundIndex ~= 0 then
					nearestVehicle = g_currentMission.steerables[nearestFoundIndex]
					if nearestVehicle.lhdX0 ~= nil then
						self.lhdX0 = nearestVehicle.lhdX0
						self.lhdZ0 = nearestVehicle.lhdZ0
						self.lhX0 = nearestVehicle.lhX0
						self.lhZ0 = nearestVehicle.lhZ0
						self.GPSWidth = nearestVehicle.GPSWidth
						self.GPS_LRoffset = nearestVehicle.GPS_LRoffset
						self.GPS_blinkTime = 2500;
						self.GPS_lastActionText = Steerable.GPS_TXT_LOAD
						needPushEvent = true;
					end
				end
				
			end
		end
		
		
		
		
		
		if InputBinding.hasEvent(InputBinding.GPS_AutoWidth) then 
			local xmin = 0;
			local xmax = 0;
			
			xmin,xmax = GPS.xMinMaxAI(self,self,xmin,xmax);
			xmin,xmax = GPS.xMinMaxAreas(self,self.cuttingAreas,xmin,xmax);
			xmin,xmax = GPS.xMinMaxAreas(self,self.mowerCutAreas,xmin,xmax);
			xmin,xmax = GPS.xMinMaxAreas(self,self.fruitPreparerAreas,xmin,xmax);

			
			for nIndex,oImplement in pairs(self.attachedImplements) do --parse implements
				if oImplement ~= nil and oImplement.object ~= nil then
					xmin,xmax = GPS.xMinMaxAI(self,oImplement.object,xmin,xmax);
					xmin,xmax = GPS.xMinMaxAreas(self,oImplement.object.cuttingAreas,xmin,xmax);
					xmin,xmax = GPS.xMinMaxAreas(self,oImplement.object.mowerCutAreas,xmin,xmax);
					xmin,xmax = GPS.xMinMaxAreas(self,oImplement.object.fruitPreparerAreas,xmin,xmax);
					for nIndex2,oImplement2 in pairs(oImplement.object.attachedImplements) do --parse 2nd line implements
						xmin,xmax = GPS.xMinMaxAI(self,oImplement2.object,xmin,xmax);
						xmin,xmax = GPS.xMinMaxAreas(self,oImplement2.object.cuttingAreas,xmin,xmax);
						xmin,xmax = GPS.xMinMaxAreas(self,oImplement2.object.mowerCutAreas,xmin,xmax);
						xmin,xmax = GPS.xMinMaxAreas(self,oImplement2.object.fruitPreparerAreas,xmin,xmax);
					end
				end
			end
			
			
			local width = math.abs(xmax-xmin);
			if width > .1 then
				self.GPSWidth = width;
				self.GPS_LRoffset = (xmin+xmax)/2;
				if math.abs(self.GPS_LRoffset) < 0.1 then
					self.GPS_LRoffset = 0;
				end
				self.GPSshowTime = 800;
				needPushEvent = true;
			end				
		end		
	end
	
	
	
	--if self:getIsActive() then
	if self:getIsActiveForInput(false) then	
		if self.GPSActive then							
			local px,py,pz = getWorldTranslation(self.rootNode);
			
			local x,y,z = localDirectionToWorld(self.rootNode, 0, 0, 1);
			local length = Utils.vector2Length(x,z);
			local dX = x/length;
			local dZ = z/length;
			
			if InputBinding.hasEvent(InputBinding.GPS_Init) or self.GPSfirstrun then
				needPushEvent = true;
				self.GPSclickTimer = 3500;
				if self.GPSangleModeLS13 then --snap to terrain directions
					local snapAngle = self:getDirectionSnapAngle();
					snapAngle = math.max(snapAngle, math.pi/(g_currentMission.terrainDetailAngleMaxValue+1));

					local angleRad = Utils.getYRotationFromDirection(x, z)
					angleRad = math.floor(angleRad / snapAngle + 0.5) * snapAngle;

					local snapX, snapZ = Utils.getDirectionFromYRotation(angleRad);	
					
					self.lhdX0 = snapX
					self.lhdZ0 = snapZ
				else -- use precisely the current orientation of the tractor. LS11 mode for non-lego field layouts
					self.lhdX0 = dX;
					self.lhdZ0 = dZ;
				end
				
				self.lhX0 = px + self.GPS_LRoffset*self.lhdZ0;
				self.lhZ0 = pz - self.GPS_LRoffset*self.lhdX0;	
				
				self.GPSfirstrun = false;
			end;
			
			
			
			local diffAnglecos = math.deg(math.acos(dX*self.lhdX0 + dZ*self.lhdZ0));
					
			local signDiffAngle = math.deg(math.asin(dX*self.lhdZ0 - dZ*self.lhdX0));
			
			local beta = 0;
					
			beta = self.lhdX0 *(pz - self.lhZ0) - self.lhdZ0 *(px - self.lhX0)
			beta = beta / self.GPSWidth		
			beta = (beta - math.floor(beta+.5)); 
			
			
			
			
			self.lhBeta = beta;
			self.lhAngle = signDiffAngle; --self.lhAngle = diffAnglecos;
			
			
			local lhDirectionPlusMinus = 1;
			if diffAnglecos > 90  then
				lhDirectionPlusMinus = -1;
			end;
			
			
			if self.GPSisActiveSteering then --here starts autopilot:
				local refangle = signDiffAngle*lhDirectionPlusMinus					
				local K1 = 15;
				local K2 = .025
				local angle_soll = K1 * (beta-lhDirectionPlusMinus*self.GPS_LRoffset/self.GPSWidth) * self.GPSWidth*lhDirectionPlusMinus
				-- local angle_soll = K1 * beta * self.GPSWidth*lhDirectionPlusMinus
				local steer = K2*(refangle - angle_soll)
				self.GPSsteeringOffset = steer
			else
				self.GPSsteeringOffset = 0									
			end
			
			local printDistance = beta * self.GPSWidth;
			printDistance = printDistance - self.GPS_LRoffset;
			if math.abs(printDistance) < 0.005 then
				printDistance = 0; --prevent +/- flickering
			end
			
			self.GPStxt.printDistance = printDistance;
			
			
			if self.GPSshowMode < 2 then --here starts line generation:
			
				local k
				local lineAx
				local lineAy
				local lineAz
				local lineBx
				local lineBy
				local lineBz
				
				local r = 0;
				local g = .8;
				local b = 0;
				local kmin = 0;
				local kmax = 30;
				local step = 2;
				
				local offsetLine = 0;
				if self.GPS_LRoffset ~= 0 then
					offsetLine = 1;
				end
				for kk = -1,1+offsetLine,1 do
				
					if kk == 0 then --middle line
						r=0;
						g=.8;
						b=0;
						kmin = 5;
						kmax = 30;
						step = 1;
					elseif math.abs(kk) == 1 then --boundary lines
						r=.8;
						g = 0;
						b=0;
						kmin = -15;
						kmax = 10;
						if not self.GPSangleModeLS13 and self.GPSclickTimer > 0 then
							kmax = 200;
						end
						step = 2;
					elseif kk == 2 then --offset line
						r=.0;
						g = 0;
						b= 0.7;
						kmin = 5;
						kmax = 30;
						step = 1.5;
					end;
					
					local line0x = 0;
					local line0z = 0;
					if kk < 2 then
						line0x = px + self.GPSWidth*self.lhdZ0*(beta+kk/2)
						line0z = pz - self.GPSWidth*self.lhdX0*(beta+kk/2)
					else
						line0x = px + self.GPSWidth*self.lhdZ0*(beta-lhDirectionPlusMinus*self.GPS_LRoffset/self.GPSWidth)
						line0z = pz - self.GPSWidth*self.lhdX0*(beta-lhDirectionPlusMinus*self.GPS_LRoffset/self.GPSWidth)					
					end
					
					local line0y = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, line0x, posy, line0z) + .2;					
						
					for k = kmin,kmax,step do
						lineAx = line0x + k*lhDirectionPlusMinus*self.lhdX0;
						lineAz = line0z + k*lhDirectionPlusMinus*self.lhdZ0;
						lineAy = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, lineAx, posy, lineAz) + .2;
						
						lineBx = line0x + (k+1)*lhDirectionPlusMinus*self.lhdX0;
						lineBz = line0z + (k+1)*lhDirectionPlusMinus*self.lhdZ0;
						lineBy = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, lineBx, posy, lineBz) + .2;
						
						drawDebugLine(lineAx, lineAy, lineAz, r, g, b, lineBx, lineBy, lineBz, r, g, b);		
					end;

				end;
			end;
		else
			self.GPSisActiveSteering= false;
		end;
	end
	
	if needPushEvent then		
        if g_server ~= nil then
            g_server:broadcastEvent(GPS_Event:new(self, self.lhX0,self.lhZ0,self.lhdX0,self.lhdZ0,self.GPSWidth,self.GPS_LRoffset), nil, nil, self);
        else
            g_client:getServerConnection():sendEvent(GPS_Event:new(self, self.lhX0,self.lhZ0,self.lhdX0,self.lhdZ0,self.GPSWidth,self.GPS_LRoffset));
        end;	
	end
end;



function GPS:draw()
	if self.GPSshowMode < 3 then
		self.GPSoverlay:render();
		
		if self.GPS_blinkTime > 0 then
			self.GPStxt.txt = self.GPS_lastActionText;
		end	
		
		if self.GPS_blinkTime == 0 or self.GPS_blinkTime % 1000 > 500 then
			setTextAlignment(RenderText.ALIGN_RIGHT); 
			setTextBold(self.GPStxt.bold);
			setTextColor(self.GPStxt.r,self.GPStxt.g,self.GPStxt.b,1);
			renderText(.3+ 0.085,0.01238 + 0.010,0.022, self.GPStxt.txt);
		end
		
		if self.GPSshowTime > 0 then
			setTextColor(1,1,1,1)
			setTextBold(false);
			renderText(.5,.8,.015,Steerable.GPS_TXT_WIDTH .. string.format(" %3.1f m", self.GPSWidth));		
			if self.GPS_LRoffset ~= 0 then
				renderText(.5,.76,.015,string.format("GPS Offset: %2.2f m", self.GPS_LRoffset));		
			end
		end;
		
		setTextAlignment(RenderText.ALIGN_LEFT); -- (hopefully) prevent influence on scripts without explicit format coding
		setTextBold(false);
		setTextColor(1,1,1,1);
	end
	
end;


function GPS:getSaveAttributesAndNodes(nodeIdent)
	local attributes = "";
	if self.lhX0 ~= nil then
		attributes = 'GPS_x="'..self.lhX0..'" '		
		attributes = attributes..'GPS_z="'..self.lhZ0..'" '
		attributes = attributes..'GPS_dx="'..self.lhdX0..'" '
		attributes = attributes..'GPS_dz="'..self.lhdZ0..'" '
		attributes = attributes..'GPS_Width="'..self.GPSWidth..'" '
		attributes = attributes..'GPS_Offset="'..self.GPS_LRoffset..'" '		
	end
	return attributes
end;


function GPS:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not resetVehicles then
		local x = getXMLFloat(xmlFile, key.."#GPS_x");
		if x ~= nil then
			self.lhX0 = x;
			
			local z = getXMLFloat(xmlFile, key.."#GPS_z");
			self.lhZ0 = z;

			local dx = getXMLFloat(xmlFile, key.."#GPS_dx");
			self.lhdX0 = dx;
			local dz = getXMLFloat(xmlFile, key.."#GPS_dz");
			self.lhdZ0 = dz;
			local w = getXMLFloat(xmlFile, key.."#GPS_Width");
			self.GPSWidth = w;
			local o = getXMLFloat(xmlFile, key.."#GPS_Offset");
			self.GPS_LRoffset = Utils.getNoNil(o,0); -- o,0 
						
			self.GPSfirstrun = false;			
		end;
	
	end;
	return BaseMission.VEHICLE_LOAD_OK;
end

function GPS:delete()	
end


function GPS.xMinMaxAI(self,object,xmin,xmax)
	
	if object.aiLeftMarker ~= nil and object.aiRightMarker ~= nil then		
		local x1,y1,z1 = getWorldTranslation(object.aiLeftMarker)
		local x2,y2,z2 = getWorldTranslation(object.aiRightMarker)
		local lx1,ly1,lz1 = worldToLocal(self.rootNode,x1,y1,z1)
		local lx2,ly2,lz2 = worldToLocal(self.rootNode,x2,y2,z2)
		
		if lx1 < xmin then
			xmin = lx1;
		end
		if lx1 > xmax then
			xmax = lx1;
		end
		if lx2 < xmin then
			xmin = lx2;
		end
		if lx2 > xmax then
			xmax = lx2;
		end
	end
	
	return xmin, xmax;
end;





function GPS.xMinMaxAreas(self,areas,xmin,xmax)

	if areas ~= nil then
		for _,cuttingArea in pairs(areas) do
			if self:getIsAreaActive(cuttingArea) then

				local x1,y1,z1 = getWorldTranslation(cuttingArea.start)
				local x2,y2,z2 = getWorldTranslation(cuttingArea.width)
				local x3,y3,z3 = getWorldTranslation(cuttingArea.height)
				local lx1,ly1,lz1 = worldToLocal(self.rootNode,x1,y1,z1)
				local lx2,ly2,lz2 = worldToLocal(self.rootNode,x2,y2,z2)
				local lx3,ly3,lz3 = worldToLocal(self.rootNode,x3,y3,z3)
				
				if lx1 < xmin then
					xmin = lx1;
				end
				if lx1 > xmax then
					xmax = lx1;
				end
				if lx2 < xmin then
					xmin = lx2;
				end
				if lx2 > xmax then
					xmax = lx2;
				end
				if lx3 < xmin then
					xmin = lx3;
				end
				if lx3 > xmax then
					xmax = lx3;
				end
			end
		end
	end

	return xmin, xmax;
end
